help = "Update/rebuild the venv"
options = [
    {
        "name": ["-g", "--global"],
        "nargs": "*",
        "dest": "global_",
        "help": "Update global packages",
        "metavar": "PACKAGE"
    }
]

def run(ns):
    from .. import venv
    if ns.global_ is None:
        update_venv(venv.get_current_venv())
        return
        
    if not ns.global_:
        ns.global_ = venv.iter_global_packages()
        
    for pkg in ns.global_:
        update_venv(venv.get_global_pkg_venv(pkg), global_pkg_name=pkg)
    
def update_venv(vv, global_pkg_name=None):
    """Update a venv.
    
    :arg vpip.venv.Venv vv: A venv instance.
    :arg str global_pkg_name: Decide how to rebuild the venv. If set then run :func:`vpip.commands.install.install_global` when rebuilding venv. Otherwise, run :func:`vpip.commands.install.install_local_first_time`
    
    If the Python version is upgraded, this command reinstall the entire venv.
    """
    import sys
    import pathlib
    import configparser
    env_dir = pathlib.Path(vv.env_dir)
    config = configparser.ConfigParser()
    config.read_string("[DEFAULT]\n" + (env_dir / "pyvenv.cfg").read_text())
    config_home = pathlib.Path(config.get("DEFAULT", "home"))
    # https://github.com/python/cpython/blob/0118d109d54bf75c99a8b0fa9aeae1a478ac4b7e/Lib/venv/__init__.py#L109
    current_home = pathlib.Path(getattr(sys, '_base_executable', sys.executable)).parent
    if config_home != current_home:
        # python is updated, rebuild the entire venv
        vv.destroy()
        from . import install
        if global_pkg_name:
            install.install_global([global_pkg_name])
        else:
            install.install_local_first_time()
        return
    
    # update pip
    from .. import pip_api
    with vv.activate():
        pip_api.install("pip", upgrade=True, latest=True)
    


help = "Link console scripts in the local venv to the global scripts folder"
options = []

def run(ns):
    from .. import venv
    vv = venv.get_current_venv()
    with vv.activate():
        link_console_script(get_current_pkg())
        
def get_current_pkg():
    from .. import dependency
    return dependency.ProdUpdater().get_name()
        
def link_console_script(pkg):
    """Find console scripts of the package and try to link the executable to
    the global scripts folder.
    
    :arg str pkg: Package name.
    """
    import shutil
    import os
    import pathlib
    from configparser import ConfigParser
    from .. import pip_api, venv
    # should be called inside a venv
    # link console script to GLOBAL_SCRIPT_FOLDER so they can be accessed outside of the venv
    entry_points = pip_api.show([pkg], verbose=True)[0].entry_points
    config = ConfigParser()
    config.read_string(entry_points)
    if "console_scripts" not in config:
        return
    for executable in config["console_scripts"]:
        src = shutil.which(executable)
        if not src:
            print("unable to access console script {}".format(executable))
            continue
        src = pathlib.Path(src)
        filename = src.name
        print("link console script '{}'".format(filename))
        
        if os.name == "nt":
            LinkerCls = WinLinker
        else:
            LinkerCls = Linker
        linker = LinkerCls(src)
        
        ok = False
        errors = []
        for folder in venv.get_global_script_folders():
            folder = pathlib.Path(folder)
            try:
                folder.mkdir(parents=True, exist_ok=True)
                linker.make(folder / filename)
            except OSError as err:
                errors.append(err)
                continue
            ok = True
            break
        if not ok:
            print("cannot link console script")
            print(errors)
        
class Linker:
    def __init__(self, src):
        self.src = src
        
    def unlink(self, dest):
        try:
            dest.unlink()
        except FileNotFoundError:
            pass
        
    def make(self, dest):
        self.unlink(dest)
        self.src.link_to(dest)

class WinLinker(Linker):
    def make(self, dest):
        # FIXME: use elevate + symlink on Windows?
        # https://stackoverflow.com/questions/6260149/os-symlink-support-in-windows
        # create a BAT file on windows
        dest = dest.with_suffix(".bat")
        self.unlink(dest)
        content = "\n".join([
            "@echo off",
            '"{}" %*'.format(self.src)
        ])
        dest.write_text(content)
        



help = "Uninstall packages and delete from dependencies"
options = [
    {
        "name": ["-g", "--global"],
        "dest": "global_",
        "action": "store_true",
        "help": "Remove packages from the global folder"
    },
    {
        "name": "PACKAGE",
        "nargs": "+",
        "help": "Package name"
    }
]

def run(ns):
    if ns.global_:
        uninstall_global(ns.PACKAGE)
    else:
        uninstall_local(ns.PACKAGE)
    
def uninstall_global(packages):
    """Uninstall global packages.
    
    :arg list[str] packages: Package names.
    """
    from .. import venv

    for pkg in packages:
        print("removing {}...".format(pkg))
        venv.get_global_pkg_venv(pkg).destroy()
    
def uninstall_local(packages):
    """Uninstall packages and remove from dependencies.
    
    :arg list[str] packages: Package names.
    """
    from .. import venv, pip_api, dependency

    vv = venv.get_current_venv()
    with vv.activate():
        pip_api.uninstall(packages)
        dependency.delete(packages)
        clean_unused()
        dependency.update_lock()

def clean_unused():
    """Remove unused packages"""
    from .. import pip_api, dependency
    deps = []
    for req in dependency.get_dev_requires():
        deps.append(req.name)
    for req in dependency.get_prod_requires():
        deps.append(req.name)
    used = set(pkg.name for pkg in pip_api.show(deps))
        
    def get_unused():
        return [
            pkg.name for pkg in pip_api.list_(not_required=True)
            if pkg.name not in used and pkg.name != "pip"
        ]
        
    unused = get_unused()
    while unused:
        pip_api.uninstall(unused)
        unused = get_unused()
        


import subprocess
import shutil

def execute(cmd, capture=False):
    """Execute a command.
    
    :arg cmd: Command. If ``cmd`` is a :class:`str`, the command would be
        invoked with shell.
    :type cmd: list or str
    :arg bool capture: If ``True`` then enter the capture mode: process output
        will be captured and the function will return a generator yielding
        lines of the output.
    :rtype: Iterator[str] or None
    """
    def do_execute():
        stdout = subprocess.PIPE if capture else None
        shell = isinstance(cmd, str)
        if not shell:
            executable = shutil.which(cmd[0])
            if executable:
                cmd[0] = executable
        with subprocess.Popen(cmd, stdout=stdout, encoding="utf8", shell=shell) as process:
            if capture:
                for line in process.stdout:
                    yield line
        if process.returncode:
            raise subprocess.CalledProcessError(process.returncode, cmd)
    if capture:
        return do_execute()
    list(do_execute())
    


from collections import namedtuple
import requests
import packaging.version

UpdateResult = namedtuple("UpdateResult", ["compatible", "latest"])

session = None

def get_session():
    """Return a static :class:`requests.Session` object used by
    :func:`check_update`, so they can share a persistent connection.
    """
    global session
    if session is None:
        session = requests.Session()
    return session

def check_update(pkg, curr_version):
    """Check update from pypi and return the result if there is an update
    available.
    
    :arg str pkg: Package name.
    :arg str curr_version: Installed version of the package.
    :rtype: UpdateResult or None
    
    :class:`UpdateResult` has two properies, ``compatible`` and ``latest``,
    which are two different version string.
    
    If ``result.compatible`` is not None, it must be compatible with
    ``curr_version`` and must lager than ``curr_version``.
    
    If ``result.latest`` is not None, it must larger than ``result.compatible``
    and ``curr_version``.
    """
    r = get_session().get("https://pypi.org/pypi/{}/json".format(pkg))
    r.raise_for_status()
    
    # curr_version = packaging.version.parse(curr_version)
    all_versions = [packaging.version.parse(v) for v in r.json()["releases"].keys()]
    all_versions = [v for v in all_versions if not v.is_prerelease]
    all_versions.sort()
    
    curr_version = packaging.version.parse(curr_version)
    latest = None
    compatible = None
    
    for version in reversed(all_versions):
        if is_compatible(curr_version, version):
            compatible = version if version != curr_version else None
            break
            
    if all_versions[-1] not in (compatible, curr_version):
        latest = all_versions[-1]
            
    if compatible or latest:
        return UpdateResult(compatible, latest)

def is_compatible(version, new_version):
    """Check if two versions are compatible. ``new_version`` may be smaller
    than ``version``.
    
    :type version: str
    :type new_version: str
    :rtype: bool
    """
    version = to_version_tuple(version)
    new_version = to_version_tuple(new_version)
    if version[0] == new_version[0]:
        if version[0] != 0:
            return True
        if version[1] == new_version[1]:
            return True
    return False
    
def to_version_tuple(version):
    """Split version into number tuple"""
    return tuple(int(n) for n in str(version).split("."))
    


from . import install

help = "Update packages to its compatible version"
options = [
    {
        "name": ["-g", "--global"],
        "dest": "global_",
        "action": "store_true",
        "help": "Update packages in the global folder"
    },
    {
        "name": ["--latest"],
        "action": "store_true",
        "help": "Install the latest version even if the new version is not compatible"
    },
    {
        "name": "PACKAGE",
        "nargs": "*",
        "help": "Package name"
    }
]

def run(ns):
    from .. import venv
    
    if ns.global_:
        if ns.PACKAGE:
            packages = ns.PACKAGE
        else:
            packages = list(venv.iter_global_packages())
        install.install_global(packages, upgrade=True, latest=ns.latest)
    else:
        update_local(
            lambda pkg: not ns.PACKAGE or pkg in ns.PACKAGE,
            latest=ns.latest)
        
def update_local(filter, latest=False):
    """Update local packages.
    
    :arg Callable filter: A ``(pkg_name: str) -> bool`` function. If return True
        then update the package.
        
        Packages would be grabbed from the dependencies.
    
    :arg bool latest: Whether to upgrade to the latest version.
    """
    from .. import dependency
    
    for requires, is_dev in [
            (dependency.get_dev_requires(), True),
            (dependency.get_prod_requires(), False)]:
        install.install_local(
            [r.name for r in requires if filter(r.name)],
            dev=is_dev,
            upgrade=True,
            latest=latest
        )
            


import re
import importlib
from pkg_resources import resource_listdir

#: List of builtin command names.
names = [
    filename.partition(".")[0]
    for filename in resource_listdir(__name__, "")
    if re.match("[a-z]\w+\.py", filename)
]

def get_modules():
    """Get module instances.
    
    :return: A ``command_name -> module`` map.
    :rtype: dict[str, module]
    """
    modules = {}
    for name in names:
        modules[name] = importlib.import_module(".{}".format(name), __name__)
    return modules
    


"""This command is just a shortcut of ``vpip list --outdated``."""

from . import list as list_

help = "List outdated dependencies"
options = [
    {
        "name": ["-g", "--global"],
        "dest": "global_",
        "action": "store_true",
        "help": "List globally installed packages"
    }
]

def run(ns):
    if ns.global_:
        list_.print_global_packages(check_outdated=True)
    else:
        list_.print_local_packages(check_outdated=True)



import os
import re
import shutil
import venv
from contextlib import contextmanager
from pathlib import Path

from .execute import execute

def get_script_folder(base):
    if os.name == "nt":
        return os.path.join(base, "Scripts")
    return os.path.join(base, "bin")
    
#: Absolute path to the global package venv folder ``~/.vpip/pkg_venvs``
GLOBAL_FOLDER = os.path.normpath(os.path.expanduser("~/.vpip/pkg_venvs"))

class GlobalScriptFolderGetter:
    """Return two folders. One is the system scripts folder and
    one is the user-site scripts folder.
    
    :rtype: list[str]
    """
    def __init__(self):
        # this includes the system folder and the user-site folder
        self.folders = None
        
    def __call__(self):
        if self.folders:
            return self.folders
        from pip._internal.locations import distutils_scheme
        self.folders = [
            # FIXME: this won't work if vpip is installed inside a venv
            distutils_scheme("foo")["scripts"],
            distutils_scheme("foo", user=True)["scripts"]
        ]
        return self.folders
        
get_global_script_folders = GlobalScriptFolderGetter()

def get_global_folder(pkg_name):
    """Get global venv folder for a package.
    
    :arg str pkg_name: Package name.
    :rtype: str
    """
    return os.path.join(GLOBAL_FOLDER, pkg_name)
    
def iter_global_packages():
    """Iterate through all venv folders for globally installed packages.
    
    :rtype: Iterator[str]
    """
    for dir in Path(GLOBAL_FOLDER).iterdir():
        yield dir.name
    
def get_path_without_venv(path, venv_dir):
    if not venv_dir:
        return path
    return os.pathsep.join(
        p for p in re.split("\s*{}\s*".format(os.pathsep), path)
        if not p.startswith(venv_dir))
                    
def get_current_venv():
    """Get the :class:`Venv` instance pointing to ``./.venv``.
    
    :rtype: Venv
    """
    return Venv(".venv")
    
def get_global_pkg_venv(pkg):
    """Get the :class:`Venv` instance pointing the venv folder of the global
    installed package.
    
    :arg str pkg: Package name. It would also be used as the folder name.
    :rtype: Venv
    """
    return Venv(get_global_folder(pkg))
    
class Builder(venv.EnvBuilder):
    """An environment builder that could be used inside a venv.
    
    It also upgrades pip to the latest version after installed.
    """
    def ensure_directories(self, env_dir):
        context = super().ensure_directories(env_dir)
        current_venv = os.environ.get("VIRTUAL_ENV")
        if current_venv:
            clean_path = get_path_without_venv(os.environ["PATH"], current_venv)
            # find executable that is not in the current virtual env
            # https://github.com/python/cpython/blob/cd449806fac1246cb7b4d392026fe6986ec01fb7/Lib/venv/__init__.py#L113-L116
            executable = shutil.which("python", path=clean_path)
            dirname, exename = os.path.split(executable)
            context.executable = executable
            context.python_dir = dirname
            context.python_exe = exename
        return context
        
    def post_setup(self, context):
        # update pip to latest
        execute([context.env_exe, "-Im", "pip", "install", "-U", "pip"])        
        

class Venv:
    """A helper class that is associated to a venv folder. It allows you to
    easily activate/deactivate the venv.
    """
    def __init__(self, env_dir):
        """
        :arg str env_dir: The target venv folder.
        """
        self.env_dir = os.path.abspath(env_dir)
        self.old_env_dir = os.environ.get("VIRTUAL_ENV")
        
        self.path = "{}{}{}".format(
            get_script_folder(self.env_dir),
            os.pathsep,
            get_path_without_venv(os.environ["PATH"], self.old_env_dir)
        )
        self.old_path = os.environ["PATH"]
        
    def exists(self):
        """Check if the folder exists.
        
        :rtype: bool
        """
        return os.path.exists(self.env_dir)
    
    @contextmanager
    def activate(self, auto_create=False):
        """Activate the venv. Update PATH and VIRTUAL_ENV environment variables.
        
        :arg bool auto_create: If True then automatically create the venv when
            the folder doesn't exist.
            
        This function can be used as a context manager that will
        :meth:`deactivate` when exited.
        """
        try:
            if not self.exists():
                if auto_create:
                    self.create()
                else:
                    raise Exception("venv folder doesn't exists")
            os.environ["PATH"] = self.path
            os.environ["VIRTUAL_ENV"] = self.env_dir
            yield
        finally:
            self.deactivate()
        
    def deactivate(self):
        """Deactivate the venv."""
        os.environ["PATH"] = self.old_path
        if self.old_env_dir:
            os.environ["VIRTUAL_ENV"] = self.old_env_dir
        elif "VIRTUAL_ENV" in os.environ:
            del os.environ["VIRTUAL_ENV"]
    
    def create(self):
        """Create the venv."""
        print("building venv at {}".format(self.env_dir))
        Builder(with_pip=True).create(self.env_dir)
        
    def destroy(self):
        """Destroy the venv. Remove the venv folder."""
        shutil.rmtree(self.env_dir)



import sys

help = "Run command inside the venv"
options = []
allow_unknown = True

def run(ns, extra):
    from subprocess import list2cmdline, CalledProcessError
    from .. import venv
    from ..execute import execute
    
    vv = venv.get_current_venv()
    with vv.activate():
        if extra and extra[0] == "--":
            extra = extra[1:]
        if not extra:
            cmd = get_shell_executable()
        elif extra == ["python"]:
            # FIXME: can't run python with shell=True?
            # https://bugs.python.org/issue35217
            cmd = extra
        else:
            cmd = list2cmdline(extra)
        try:
            execute(cmd)
        except CalledProcessError as err:
            sys.exit(err.returncode)
    
def get_shell_executable():
    # FIXME: this only works on Windows
    return ["cmd", "/k"]
    


import configparser
import re
from pathlib import Path

from configupdater import ConfigUpdater
from pkg_resources import parse_requirements

LOCK_FILE = "requirements-lock.txt"

def get_dev_requires():
    return parse_requirements(DevUpdater().get_requirements())
    
def get_prod_requires():
    return parse_requirements(ProdUpdater().get_requirements())
    
class Updater:
    """Dependency updater interface. Extend this class to create a new updater.
    """
    def get_requirements(self):
        """Get requirements string.
        
        :rtype: str
        """
        raise NotImplementedError
        
    def get_spec(self, name, version):
        """Get version specifier.
        
        :arg str name: Installed package name.
        :arg str version: Installed pacakge version.
        :return: Version specifier e.g. ``"foo==0.1.0"``
        :rtype: str
        """
        raise NotImplementedError
        
    def write_requirements(self, lines):
        """Write new requirements to file.
        
        :arg list[str] line: Lines of requirements.
        """
        raise NotImplementedError

class DevUpdater(Updater):
    """Development dependency (requirements.txt) updater."""
    def __init__(self):
        self.file = Path("requirements.txt")
        
    def get_requirements(self):
        try:
            return self.file.read_text("utf8")
        except OSError:
            pass
        return ""
        
    def get_spec(self, name, version):
        return "{}=={}".format(name, version)
        
    def write_requirements(self, lines):
        with self.file.open("w", encoding="utf8") as f:
            for line in lines:
                f.write(line + "\n")
                
class ProdUpdater(Updater):
    """Production dependency (setup.cfg) updater."""
    def __init__(self):
        self.file = Path("setup.cfg")
        self.config = ConfigUpdater()
        self.indent = None
        
    def read(self):
        try:
            text = self.file.read_text("utf8")
            self.indent = detect_indent(text)
            self.config.read_string(text)
        except OSError:
            pass        
        
    def get_requirements(self):
        self.read()
        try:
            return self.config.get("options", "install_requires").value
        except configparser.Error:
            pass
        return ""
        
    def get_name(self):
        self.read()
        return self.config.get("metadata", "name").value
        
    def get_spec(self, name, version):
        if not version.startswith("0."):
            version = re.match("\d+\.\d+", version).group()
        return "{}~={}".format(name, version)
        
    def write_requirements(self, lines):
        if "options" not in self.config:
            self.config.add_section("options")
        self.config.set("options", "install_requires", "".join(
            "\n" + self.indent + l for l in lines))
        self.file.write_text(str(self.config).replace("\r", ""), "utf8")
        
class UpdateDependencyResult:
    def __init__(self):
        self.dirty = False
        self.add = 0
        self.update = 0
        self.remove = 0
        self.incompat_update = 0
        
def update_dependency(updater, added=None, removed=None):
    """Update dependency and save.
    
    :arg Updater updater: An Updater instance.
    :arg dict added: A ``pkg_name -> version`` map. Added packages.
    :arg list[str] removed: A list of package name. Removed packages.
    """
    added = added or {}
    removed = set(removed or [])
    output = []
    result = UpdateDependencyResult()
    for require in parse_requirements(updater.get_requirements()):
        if require.name in added:
            result.dirty = True
            result.update += 1
            version = added.pop(require.name)
            if version not in require.specifier:
                result.incompat_update += 1
            spec = updater.get_spec(require.name, version)
            if require.marker:
                spec += ";{}".format(require.marker)
            output.append(spec)
        elif require.name in removed:
            result.dirty = True
            result.remove += 1
        else:
            output.append(str(require))
            
    for name, version in added.items():
        result.dirty = True
        result.add += 1
        output.append(updater.get_spec(name, version))
        
    if result.dirty:
        output.sort()
        updater.write_requirements(output)
        
    return result
        
def has_lock():
    """Detect if there is a lock file (requirements-lock.txt)"""
    return Path(LOCK_FILE).exists()
        
def update_lock():
    """Run ``pip freeze`` and update the lock file"""
    from . import pip_api
    lines = []
    for pkg in pip_api.list_():
        if pkg.name == "pip":
            continue
        lines.append("{}=={}".format(pkg.name, pkg.version))
    Path(LOCK_FILE).write_text("\n".join(lines))

def add_dev(packages):
    return update_dependency(DevUpdater(), added=packages)
    
def add_prod(packages, **kwargs):
    return update_dependency(ProdUpdater(), added=packages)

def delete(packages):
    return (
        update_dependency(DevUpdater(), removed=packages),
        update_dependency(ProdUpdater(), removed=packages)
    )
    
def detect_indent(text):
    for line in text.split("\n"):
        match = re.match("(\s+)\S", line)
        if match:
            return match.group(1)
    return None
    


"""``pip`` command API."""

import json
import re
from argparse import Namespace
from subprocess import CalledProcessError

import case_conversion
from pkg_resources import Requirement

from .execute import execute

def install(package, install_scripts=None, upgrade=False, latest=False):
    """Install a package and return the package info.
    
    :arg str package: Package name. It may include the version specifier.
    :arg str install_scripts: Install scripts to a different folder. It uses
        the ``--install-option="--install-scripts=..."`` pip option.
    :arg bool upgrade: Upgrade package.
    :arg bool latest: Whether upgrade to the latest version. Otherwise upgrade
        to the compatible version. This option has no effect if ``package``
        includes specifiers.
    :return: Package information returned by :func:`show`.
    :rtype: Namespace
    """
    cmd = "install"
    require = Requirement.parse(package)
    if install_scripts:
        cmd += " --install-option \"--install-scripts={}\"".format(install_scripts)
    if upgrade:
        cmd += " -U"
        if not latest and not require.specs:
            try:
                version = show([require.name])[0].version
            except CalledProcessError:
                pass
            else:
                package = "{}~={}".format(require.name, get_compatible_version(version))
    execute_pip("{} {}".format(cmd, package))
    return show([require.name])[0]
    
def install_requirements(file="requirements.txt"):
    """Install ``requirements.txt`` file."""
    execute_pip("install -r {}".format(file))
    
def install_editable():
    """Install the current cwd as editable package."""
    execute_pip("install -e .")
    
def uninstall(packages):
    """Uninstall packages.
    
    :arg list[str] package: Package name.
    """
    execute_pip("uninstall -y {}".format(" ".join(packages)))
    
def show(packages, verbose=False):
    """Get package information.
    
    :arg list[str] packages: A list of package name.
    :arg bool verbose: Whether to return verbose info.
    :return: A list of namespace objects holding the package information.
    :rtype: list[Namespace]
    
    This function uses ``pip show`` under the hood. Property name is generated
    by :func:`case_conversion.snakecase`.
    """
    cmd = "show"
    if verbose:
        cmd += " --verbose"
        
    result = []
    ns = Namespace()
    last_name = None
    
    for line in execute_pip("{} {}".format(cmd, " ".join(packages)), True):
        if line.startswith("---"):
            result.append(ns)
            ns = Namespace()
            continue
            
        match = re.match("([\w-]+):\s*(.*)", line)
        if match:
            name, value = match.groups()
            name = case_conversion.snakecase(name)
            value = value.strip()
            setattr(ns, name, value)
            last_name = name
            continue
            
        match = re.match("\s+(\S.*)", line)
        if match and last_name:
            value = getattr(ns, last_name) + "\n" + match.group(1).strip()
            setattr(ns, last_name, value)
            continue
            
    result.append(ns)
    return result
    
def list_(not_required=False, format="json"):
    """List installed packages.
    
    :rtype: list[argparse.Namespace]
    """
    cmd = "list --local --exclude-editable"
    if not_required:
        cmd += " --not-required"
    cmd += " --format {}".format(format)
    lines = []
    for line in execute_pip(cmd, capture=True):
        lines.append(line)
    return [create_ns_from_dict(item) for item in json.loads("".join(lines))]
    
def create_ns_from_dict(d):
    """Create a namespace object from a dict.
    
    :arg dict d: Dictionary.
    :rtype: argparse.Namespace
    """
    ns = Namespace()
    for key, value in d.items():
        setattr(ns, key, value)
    return ns

def execute_pip(cmd, capture=False):
    """Run pip command.
    
    :arg str cmd: ``pip`` command. It would be prefixed with ``python -m pip``.
    :arg bool capture: Whether to capture output.
    """
    prefix = "python -m pip "
    if capture:
        prefix += "--no-color "
    return execute(prefix + cmd, capture)
    
def get_compatible_version(version):
    """Return the compatible version.
    
    :arg str version: Version string.
    :return: The compatible version which could be used as ``~={compatible_version}``.
    :rtype: str

    Suppose the version string is ``x.y.z``:
    
    * If ``x`` is zero then return ``x.y.z``.
    * Otherwise return ``x.y``.
    """
    if version.startswith("0."):
        return version
    return ".".join(version.split(".")[:2])
    


from .link import link_console_script

help = "Install package and save to dependencies"
options = [
    {
        "type": "exclusive_group",
        "options": [
            {
                "name": ["-g", "--global"],
                "dest": "global_",
                "action": "store_true",
                "help": "Install packages to the global folder"
            },
            {
                "name": ["-D", "--save-dev"],
                "action": "store_true",
                "help": "Save to development dependencies (requirements.txt) "
                        "instead of production dependencies (setup.cfg)"
            }
        ]
    },
    {
        "name": "PACKAGE",
        "nargs": "*",
        "help": "Package name"
    }
]

def run(ns):
    if ns.global_:
        if not ns.PACKAGE:
            raise Exception("no package to install to global folder")
        install_global(ns.PACKAGE)
    elif ns.PACKAGE:
        install_local(ns.PACKAGE, dev=ns.save_dev)
    else:
        install_local_first_time()

def install_global(packages, upgrade=False, latest=False):
    """Install global packages.
    
    :arg list[str] packages: List of package name.
    :arg bool upgrade: Upgrade package. By default, the function skipped
        already installed packages.
    :arg bool latest: Upgrade to the latest version. By default, only
        compatible versions are selected.
    """
    from .. import venv, pip_api
    for pkg in packages:
        vv = venv.get_global_pkg_venv(pkg)
        if vv.exists() and not upgrade:
            print("{} is already installed".format(pkg))
            continue
        try:
            with vv.activate(True):
                # TODO: make pip support install_scripts
                # https://github.com/pypa/pip/issues/3934
                pip_api.install(pkg, upgrade=upgrade, latest=latest)
                link_console_script(pkg)
        except Exception:
            vv.destroy()
            raise

def install_local(packages, dev=False, **kwargs):
    """Install local packages and save to dependency.
    
    :arg list[str] packages: List of package name.
    :arg bool dev: If true then save to development depedency. Otherwise, save
        to production dependency.
    :arg dict kwargs: Other arguments are sent to :func:`vpip.pip_api.install`.
    """
    if not packages:
        return
        
    from .. import venv, pip_api, dependency
    vv = venv.get_current_venv()
    installed = {}
    with vv.activate(True):
        for pkg in packages:
            result = pip_api.install(pkg, **kwargs)
            installed[pkg] = result.version
        if dev:
            dependency.add_dev(installed)
        else:
            result = dependency.add_prod(installed)
            if result.incompat_update:
                # rebuild egg file to avoid incompatible errors
                # https://github.com/eight04/vpip/issues/19
                pip_api.install_editable()
        dependency.update_lock()

def install_local_first_time():
    """Create the venv and install all dependencies.
    
    If the lock file exists, execute ``pip install -r requirements-lock.txt``.
    
    Otherwise ``pip install -e . && pip install -r requirements.txt``.
    """
    from .. import venv, pip_api, dependency
    vv = venv.get_current_venv()
    with vv.activate(True):
        if dependency.has_lock():
            pip_api.install_requirements(dependency.LOCK_FILE)
            pip_api.install_editable()
        else:
            pip_api.install_editable()
            pip_api.install_requirements()
            dependency.update_lock()



help = "List dependencies"
options = [
    {
        "name": ["-g", "--global"],
        "dest": "global_",
        "action": "store_true",
        "help": "List globally installed packages"
    },
    {
        "name": ["--outdated"],
        "action": "store_true",
        "help": "List outdated packages only"
    }
]

def run(ns):
    if ns.global_:
        print_global_packages(check_outdated=ns.outdated)
    else:
        print_local_packages(check_outdated=ns.outdated)
        
def iter_global_packages():
    """Iterate through globally installed packages.
    
    :rtype: Iterator[PackageInfo]
    """
    from .. import venv, pip_api
    for name in venv.iter_global_packages():
        vv = venv.get_global_pkg_venv(name)
        with vv.activate():
            yield PackageInfo(name, pip_api.show([name])[0].version)
            
def print_global_packages(check_outdated=False):
    for info in iter_global_packages():
        if check_outdated and not info.check_update():
            continue
        print(info)
        
def print_local_packages(check_outdated=False):
    from packaging.utils import canonicalize_name
    from .. import venv, pip_api, dependency
    vv = venv.get_current_venv()
    dev_requires = list(dependency.get_dev_requires())
    prod_requires = list(dependency.get_prod_requires())
    
    installed = {}
    with vv.activate():
        for info in pip_api.list_():
            installed[canonicalize_name(info.name)] = info.version
            
    def get_infos(requires):
        for require in requires:
            info = PackageInfo(
                require.name,
                installed.get(canonicalize_name(require.name))
            )
            if check_outdated and not info.check_update():
                continue
            yield info
        
    dev_count = 0
    for info in get_infos(dev_requires):
        if not dev_count:
            print("-- Dev dependency --")
        print(str(info))
        dev_count += 1
    
    prod_count = 0
    for info in get_infos(prod_requires):
        if dev_count and not prod_count:
            print("")
        if not prod_count:
            print("-- Prod dependency --")
        print(str(info))
        prod_count += 1

class PackageInfo:
    """Package information formatter.
    
    Usage::
        
        print(PackageInfo("my_package", "0.1.0"))
    """
    def __init__(self, name, version=None):
        """
        :arg str name: Package name.
        :arg str version: Package version. If none then it shows
            ``(not installed)`` when formatted.
        """
        #: 
        self.name = name
        #: 
        self.version = version
        #: Update result. This would be set to the return value of
        #: :func:`vpip.pypi.check_update`.
        self.update_result = None
            
    def check_update(self):
        """Check update and setup :attr:`update_result`."""
        if not self.version:
            return False
        from .. import pypi
        self.update_result = pypi.check_update(self.name, self.version)
        return self.update_result is not None
            
    def __str__(self):
        """Format package information."""
        lines = []
        if self.version:
            lines.append("{} {}".format(self.name, self.version))
            if self.update_result:
                if self.update_result.compatible:
                    lines.append("  Update: {} (compatible)".format(self.update_result.compatible))
                if self.update_result.latest:
                    lines.append("  Update: {} (latest)".format(self.update_result.latest))
        else:
            lines.append("{} (not installed)".format(self.name))
        return "\n".join(lines)
        


import argparse
import sys
import shlex
from configparser import ConfigParser
from pathlib import Path

from . import commands

def cli(args=None):
    """CLI entry point.
    
    :arg list args: Argument list. Use ``sys.argv[1:]`` if None.
    
    You can invoke this function like::
    
        from vpip.cli import cli
        cli(["install", "-g", "my-package"])
    """
    patch_argparse()
        
    parser = argparse.ArgumentParser(prog="vpip", description="A CLI which aims to provide an npm-like experience when installing Python packages.")
    
    if args is None:
        args = sys.argv[1:]
    
    def fallback(values):
        config = ConfigParser()
        file = Path("setup.cfg")
        try:
            config.read_string(file.read_text(encoding="utf8"))
            if "vpip.commands" in config:
                for key, value in config["vpip.commands"].items():
                    if key == values[0]:
                        return ["run", *shlex.split(value), *values[1:]]
            fallback_command = config.get("vpip", "command_fallback", fallback=None)
            if fallback_command is not None:
                return ["run", *shlex.split(fallback_command), *values]
        except OSError:
            pass
            
    subparsers = parser.add_subparsers(
        title="subcommands", dest="COMMAND", required=True, metavar="COMMAND",
        fallback=fallback)
        
    modules = commands.get_modules()
    for name, module in modules.items():
        command = subparsers.add_parser(
            name, help=module.help, description="{}.".format(module.help))
        add_arguments(command, getattr(module, "options"))
        
    ns, extra = parser.parse_known_args(args)
    
    module = modules[ns.COMMAND]
    if getattr(module, "allow_unknown", False):
        module.run(ns, extra)
    elif not extra:
        module.run(ns)
    else:
        parser.error('unreconized arguments: {}'.format(' '.join(extra)))
    
def add_arguments(parser, options):
    """Add JSON-formatted argument list to parser.
    
    :arg argparse.ArgumentParser parser: The parser, or anything that
        implements ``add_argument`` method.
    :arg list options: List of options. See the source code in `vpip.commands <https://github.com/eight04/vpip/blob/95c9e03acf8239b342759aab238b28591cd4f214/vpip/commands/install.py#L2>`_ for example.
    """
    for option in options:
        type = option.pop("type", None)
        if type == "exclusive_group":
            sub_options = option.pop("options", [])
            group = parser.add_mutually_exclusive_group(**option)
            add_arguments(group, sub_options)
        else:
            name = option.pop("name")
            if not isinstance(name, list):
                name = [name]
            parser.add_argument(*name, **option)
            
is_argparse_patched = False

def patch_argparse():
    """Patch :mod:`argparse`. Make :meth:`ArgumentParser.add_subparsers`
    accept a new keyword argument ``fallback``, which is a function receving a
    ``args`` list and returning a new ``args`` or None.
    
    This function can be called multiple times, but only the first call will
    patch the module.
    """
    global is_argparse_patched
    if is_argparse_patched:
        return
    is_argparse_patched = True
    
    # pylint: disable=protected-access
    __init__ = argparse._SubParsersAction.__init__
    def init(self, *args, **kwargs):
        self._fallback = kwargs.pop("fallback", None)
        __init__(self, *args, **kwargs)
        if self._fallback:
            self.choices = None
    argparse._SubParsersAction.__init__ = init
    
    __call__ = argparse._SubParsersAction.__call__
    def call(self, *args, **kwargs):
        values = args[2]
        if values[0] not in self._name_parser_map and self._fallback:
            new_values = self._fallback(values)
            if new_values is not None:
                values[:] = new_values
        return __call__(self, *args, **kwargs)
    argparse._SubParsersAction.__call__ = call
